/* PLABA 1.0
*
================================================================================
*
*   Special version of the SOUND programme which only does some playbacking.
*   (It takes a sampled audio-file and plays it at the specified volume.)
*
================================================================================
*
*   Name of programme         : PLABA 1.0
*
*   Author                    : Maximilian Hohmann (and SUN Computers ...)
*   Date of first programming : January, 1990
*   Date of last modification : January 31, 1990
*   Purpose                   : Play a soundfile on a SPARC-1 station.
*   Use                       : plaba soundfile [-l] [-v<volume>]
*   Specialities              : If soundfile is 'jamais.au', then some lines
*                               of text will be displayed on the screen !
*   Variables                 : They are not explained at all ...
*   External functions        : So far none required.
*   Compilation notes         : cc -O -sun4 plaba.c -o plaba
*
*   Modifications 	      : by G.Lekkas, April 1990.
*				made it lint-free, added continuous looping,
*				changed args to Unix-style , removed 'dir' argument
*
*/

#include <sys/types.h>
#include <sys/dir.h>
#include <sys/file.h>

#include <fcntl.h>
#include <errno.h>
#include <stdio.h>

#include <sys/ioctl.h>
#include <sun/audioio.h>

#define	FPRINTF	(void) fprintf
#define	SPRINTF	(void) sprintf
#define	STRCPY	(void) strcpy

#define WRITE_SIZE      4096
#define BUFFER_SIZE	819200
#define WAVEFORM_WIDTH	1600
#define TEST_DEVICE	"audio"

#define ALERT_FILE	"/usr/demo/SOUND/sounds/alert.au"

int                 Audio_fd;
int                 Loop_flag;

#if defined(DEBUG) || defined(lint)
int                 Real_flag;

#else
#define	Real_flag 1
#endif

extern int          errno;
extern char        *sys_errlist[];

void                describe_proc ();
void                control_output_proc ();
void                control_loop_proc ();

char               *strcpy ();
char               *sprintf ();

int                 Compress = 1;

#define MODULATION_U_LAW	1
#define MODULATION_A_LAW	2
#define MODULATION_BINARY	3

struct sound_buffer {
	int                 bits_per_sample;
	int                 samples_per_second;
	int                 channels;
	int                 modulation;
	int                 allocated;
	int                 size;
	int                 cursor_1;
	int                 cursor_2;
	int                 cursor_start;
	int                 cursor_end;
	int                 display_position;
	int                 last_displayed;
	int                 io_position;
	char                data[BUFFER_SIZE];
}                   Buffer;

char                filename[80];
char                lohengrin[80];
int                 file_end;

/* *******************************************************************
   *** HERE WE GO ! **************************************************
   *******************************************************************
                                                                    */

main (argc, argv)
	int                 argc;
	char               *argv[];

{
	int                 volume;

	int                 bytes;
	int                 start;
	int                 end;
	int                 queue_size;
	int                 rtn;
	int                 text_displayed = 0;

	int                 fd;
	char                message[256];
	char                path[256];

	char                buffer[8192];

	void                main_play_volume_proc ();

	if (argc < 2) {
		printf ("\n\n Dayah wanna drive me mad or summpin ?\n");
		printf (" Ya know that I get called with \"%s filename [-l] [-v<volume>]\" !\n", argv[0]);
		printf (" Where  filename    = Digitized sound file\n");
		printf ("        -l 	   = Loop continuously\n");
		printf ("        -v<volume>  = Playback volume (0 ... 20)\n");
		printf (" Try again later or leave it all at once !\n\n");
		exit (1);
	}
	strcpy (filename, argv[1]);
	strcpy (lohengrin, "jamais.au");

	while (argc-- > 2)
		switch (argv[argc][1]) {
		    case 'l':
			Loop_flag = 1;
			break;
		    case 'v':
			volume = atoi (argv[argc] + 2) - 10;
			break;
		    default:
			break;
		}

	audio_open ();

	buffer_initialize (&Buffer);

	main_play_volume_proc (volume);

	file_end = 0;

	/* Starts with the contents of file_load_proc                            */

	SPRINTF (path, "%s", filename);

	if (path[0] == '\0') {
		message_display (ALERT_FILE);
		exit (1);
	}
	if ((fd = open (path, 0)) < 0) {
		SPRINTF (path, "/usr/demo/SOUND/sounds/%s", filename);
		fd = open (path, 0);
	}
	if (fd < 0) {
		SPRINTF (message, "Can't read '%s' (%s).", filename,
			 sys_errlist[errno]);
		message_display (ALERT_FILE);
		exit (1);
	}
	/* Get all the necessary information about the audio file            */
	Buffer.size = read (fd, Buffer.data, BUFFER_SIZE);
	Buffer.display_position = 0;
	Buffer.samples_per_second = 8192;
	Buffer.bits_per_sample = 8;
	Buffer.channels = 1;
	Buffer.modulation = MODULATION_U_LAW;
	Buffer.cursor_1 = 0;
	Buffer.cursor_2 = Buffer.size - 1;
	Compress = 1;

	/*
	 * Calculate the cursor positions every time we enter this loop. This
	 * will let us change the cursors on the fly and have them take effect
	 * immediately.
	 */

	if (Buffer.cursor_1 < Buffer.cursor_2) {
		Buffer.cursor_start = Buffer.cursor_1;
		Buffer.cursor_end = Buffer.cursor_2;
		Buffer.io_position = Buffer.cursor_start;
	} else {
		Buffer.cursor_start = Buffer.cursor_2;
		Buffer.cursor_end = Buffer.cursor_1;
		Buffer.io_position = Buffer.cursor_start;
	}

play:				/* Now comes main_play_proc */

	/*
	 * See if there is enough room in the device for another block of
	 * sound.  We want to write in large chunks and keep as much data
	 * enqueued as possible.
	 */

	if (Real_flag) {
	} else {
		bytes = 0;
		queue_size = 16384;
	}

	start = Buffer.io_position;
	end = start + WRITE_SIZE;
	if (((queue_size - bytes) > WRITE_SIZE) &&
	    (start < Buffer.cursor_end)) {
		if (end > Buffer.cursor_end) {
			end = Buffer.cursor_end;
		}
		Buffer.io_position = end;

		if (Real_flag) {
			rtn = write (Audio_fd, &Buffer.data[start],
				     end - start);
			if (rtn < 0) {
				FPRINTF (stderr,
					 "write returned %d (%s)\n",
					 rtn, sys_errlist[errno]);
			}
		}
	}
	if (Real_flag) {
	}

	/*
	 * XXX We don't want to return until we are done playing the sound.
	 */

	if ((end >= Buffer.cursor_end) && (bytes == 0)) {
		if (Loop_flag == 0) {

			goto bye;
		} else {
			Buffer.io_position = Buffer.cursor_start;
			Buffer.last_displayed = -1;
		}
	}
	bytes = 8192;

	/* Read in the next bit of a buffer ...                                  */
	rtn = read (fd, buffer, 8192);
	if (rtn < 0) {
		printf ("rtn < 0 while reading\n");
		goto bye;
	}
	/* It may display some lines of text now ...                           */
	if (text_displayed == 0) {
		if (!strcmp (lohengrin, filename)) {	/* Display some lines of
							 * test here */
			(void) system ("clear");
			printf ("\n\n\n");
			printf ("Lohengrin: Nie sollst du mich befragen,\n");
			printf ("           Noch wissend Sorge tragen\n");
			printf ("           woher ich kam der Fahrt,\n");
			printf ("           noch wie mein Nam' und Art !\n");
			printf ("\n");
			printf ("Elsa:      Nie, Herr, soll mir die Frage kommen\n");
			printf ("\n");
			printf ("Lohengrin: EEEELSA ! Hast du mich wohl vernommen:\n");
			printf ("           Nie sollst du mich befragen,\n");
			printf ("           Noch wissend Sorge tragen,\n");
			printf ("           woher ich kam der Fahrt,\n");
			printf ("           noch wie mein Nam' und Art !\n");
			printf ("\n\n\n");
		}
		text_displayed = 1;
	}
	goto play;

bye:				/* EOF reached, it's all over now, baby blue */

	(void) close (fd);
	exit (0);

}

/*  The play volume can be controlled by both the GR and GER registers.  To get
 *  a larger range of control, we use both registers.  For example, if we select
 *  a volume of 2, then we set each register to 1.  If we set a volume of 3, we
 *  set one register to 1 and the other to 2.  The sum of the GR and GER values
 *  is the desired volume.
 */

void 
main_play_volume_proc (value)
	int                 value;

{
	int                 gr;
	int                 ger;

	ger = value / 2;
	gr = value - ger;
	if (ger < -10) {
		ger = -10;
		gr = value - ger;
	}
	if (gr > 12) {
		gr = 12;
		ger = value - gr;
	}
	if ((gr + ger) != value) {
		FPRINTF (stderr,
			 "Error: gr (%d) + ger (%d) != value (%d)\n",
			 gr, ger, value);
	}
	audio_set_gr (gr);
	audio_set_ger (ger);
}

/***************************************************************************
	Audio related functions.
*/

audio_open ()
{

#ifndef Real_flag
	Real_flag = 1;
#endif

	if ((Audio_fd = open ("/dev/audio", 2)) < 0) {
#ifdef DEBUG
		FPRINTF (stderr,
			 "Can't open /dev/audio -- using file named %s\n",
			 TEST_DEVICE);
		Real_flag = 0;
		if ((Audio_fd = open (TEST_DEVICE, 2)) < 0) {
			perror (TEST_DEVICE);
			exit (1);
		}
#else DEBUG
		perror ("/dev/audio");
		exit (1);
#endif DEBUG
	}
	/* Initialize the three gain registers to 0db. */
	audio_set_ger (0);
	audio_set_gr (0);
	audio_set_gx (0);

	/*
	 * Tell the chip to set the gains according to the register values we
	 * just set.
	 */



	/*
	 * Initialize the MMR2 register to send the output to the builtin
	 * speaker.  This is the default, but we will be defensive anyway.
	 * Note that we read the register and turn on the appropriate bit,
	 * rather thanjust setting it.  This keeps us from stompin any useful
	 * bits that are already set.  It turns out that there is one -- the
	 * bit that selects AINB for recording.
	 */
}

 /*
  * These are tables of values to be loaded into various gain registers.
  * 
  * Note that for the ger entry for -8db, we use the data sheet value for -7.5db.
  * The data sheet gives values for -8db which are wrong and produce too much
  * gain.
  */

static unsigned char ger_table[][2] = {
	0xaa, 0xaa,		/* -10db */
	0x79, 0xac,
	/* 0x41,	0x91, */
	0x31, 0x99,		/* -7.5db */
	0x9c, 0xde,
	0x74, 0x9c,		/* -6db */
	0x6a, 0xae,
	0xab, 0xdf,
	0x64, 0xab,
	0x2a, 0xbd,
	0x5c, 0xce,
	0x00, 0x99,		/* 0db */
	0x43, 0xdd,
	0x52, 0xef,
	0x55, 0x42,
	0x31, 0xdd,
	0x43, 0x1f,
	0x40, 0xdd,		/* 6db */
	0x44, 0x0f,
	0x31, 0x1f,
	0x10, 0xdd,
	0x41, 0x0f,
	0x60, 0x0b,
	0x42, 0x10,		/* 12db */
	0x11, 0x0f,
	0x72, 0x00,
	0x21, 0x10,
	0x22, 0x00,
	0x00, 0x0b,
	0x00, 0x0f,		/* 18db */
};

static unsigned char gr_gx_table[][2] = {
	0x8b, 0x7c,		/* -18db */
	0x8b, 0x35,
	0x8b, 0x24,
	0x91, 0x23,
	0x91, 0x2a,
	0x91, 0x3b,
	0x91, 0xf9,		/* -12db */
	0x91, 0xb6,
	0x91, 0xa4,
	0x92, 0x32,
	0x92, 0xaa,
	0x93, 0xb3,
	0x9f, 0x91,		/* -6db */
	0x9b, 0xf9,
	0x9a, 0x4a,
	0xa2, 0xa2,
	0xaa, 0xa3,
	0xbb, 0x52,
	0x08, 0x08,		/* 0db */
	0x3d, 0xac,
	0x25, 0x33,
	0x21, 0x22,
	0x12, 0xa2,
	0x11, 0x3b,
	0x10, 0xf2,		/* 6db */
	0x02, 0xca,
	0x01, 0x5a,
	0x01, 0x12,
	0x00, 0x32,
	0x00, 0x13,
	0x00, 0x0e,		/* 12db */
};

audio_set_ger (value)
	int                 value;
{

	if ((value < -10) || (value > 18)) {
		FPRINTF (stderr,
			 "GER value %d out of range; %d <= GER <=  %d\n",
			 value, 0, 18);
		return;
	}

	/*
	 * Add 10 to the value to get the index into the table.
	 */

	if (Real_flag) {
	}
}

audio_set_gr (value)
	int                 value;
{

	if ((value < -18) || (value > 12)) {
		FPRINTF (stderr,
			 "GR value %d out of range; %d <= GR <=  %d\n",
			 value, 0, 12);
		return;
	}

	if (Real_flag) {
	}
}

audio_set_gx (value)
	int                 value;
{

	if ((value < -18) || (value > 12)) {
		FPRINTF (stderr,
			 "GX value %d out of range; %d <= GX <=  %d\n",
			 value, 0, 12);
		return;
	}

	/*
	 * We add 18 to get the index into the table, since entry 0 represents
	 * -18db.
	 */

	if (Real_flag) {
	}
}

play_file (filename)
	char                filename[];

{
	int                 fd;
	int                 rtn;
	char                buffer[8192];

	if ((fd = open (filename, 0)) < 0) {
		return;
	}
	rtn = read (fd, buffer, 8192);
	if (rtn < 0) {
		return;
	}
	if (write (Audio_fd, buffer, rtn) < 0) {
		perror ("/dev/audio");
	}
	(void) close (fd);

}

buffer_initialize (buffer)
	struct sound_buffer *buffer;
{
	extern char        *getwd ();

	buffer -> size = 0;
	buffer -> cursor_1 = 0;
	buffer -> cursor_2 = 0;

	bzero (buffer -> data, buffer -> size);

}

message_display (file)
	char                file[];

{
	if (file[0]) {
		play_file (file);
	}
}
